dual_fund_csv = [
    "msgtype,tx_add_input,66",
    "msgdata,tx_add_input,channel_id,channel_id,",
    "msgdata,tx_add_input,serial_id,u64,",
    "msgdata,tx_add_input,prevtx_len,u16,",
    "msgdata,tx_add_input,prevtx,byte,prevtx_len",
    "msgdata,tx_add_input,prevtx_vout,u32,",
    "msgdata,tx_add_input,sequence,u32,",
    "msgdata,tx_add_input,script_sig_len,u16,",
    "msgdata,tx_add_input,script_sig,byte,script_sig_len",
    "msgtype,tx_add_output,67",
    "msgdata,tx_add_output,channel_id,channel_id,",
    "msgdata,tx_add_output,serial_id,u64,",
    "msgdata,tx_add_output,sats,u64,",
    "msgdata,tx_add_output,scriptlen,u16,",
    "msgdata,tx_add_output,script,byte,scriptlen",
    "msgtype,tx_remove_input,68",
    "msgdata,tx_remove_input,channel_id,channel_id,",
    "msgdata,tx_remove_input,serial_id,u64,",
    "msgtype,tx_remove_output,69",
    "msgdata,tx_remove_output,channel_id,channel_id,",
    "msgdata,tx_remove_output,serial_id,u64,",
    "msgtype,tx_complete,70",
    "msgdata,tx_complete,channel_id,channel_id,",
    "msgtype,tx_signatures,71",
    "msgdata,tx_signatures,channel_id,channel_id,",
    "msgdata,tx_signatures,txid,sha256,",
    "msgdata,tx_signatures,num_witnesses,u16,",
    "msgdata,tx_signatures,witness_stack,witness_stack,num_witnesses",
    "subtype,witness_stack",
    "subtypedata,witness_stack,num_input_witness,u16,",
    "subtypedata,witness_stack,witness_element,witness_element,num_input_witness",
    "subtype,witness_element",
    "subtypedata,witness_element,len,u16,",
    "subtypedata,witness_element,witness,byte,len",
    "msgtype,open_channel2,64",
    "msgdata,open_channel2,chain_hash,chain_hash,",
    "msgdata,open_channel2,channel_id,channel_id,",
    "msgdata,open_channel2,funding_feerate_perkw,u32,",
    "msgdata,open_channel2,commitment_feerate_perkw,u32,",
    "msgdata,open_channel2,funding_satoshis,u64,",
    "msgdata,open_channel2,dust_limit_satoshis,u64,",
    "msgdata,open_channel2,max_htlc_value_in_flight_msat,u64,",
    "msgdata,open_channel2,htlc_minimum_msat,u64,",
    "msgdata,open_channel2,to_self_delay,u16,",
    "msgdata,open_channel2,max_accepted_htlcs,u16,",
    "msgdata,open_channel2,locktime,u32,",
    "msgdata,open_channel2,funding_pubkey,point,",
    "msgdata,open_channel2,revocation_basepoint,point,",
    "msgdata,open_channel2,payment_basepoint,point,",
    "msgdata,open_channel2,delayed_payment_basepoint,point,",
    "msgdata,open_channel2,htlc_basepoint,point,",
    "msgdata,open_channel2,first_per_commitment_point,point,",
    "msgdata,open_channel2,channel_flags,byte,",
    "msgdata,open_channel2,tlvs,opening_tlvs,",
    "tlvtype,opening_tlvs,option_upfront_shutdown_script,1",
    "tlvdata,opening_tlvs,option_upfront_shutdown_script,shutdown_len,u16,",
    "tlvdata,opening_tlvs,option_upfront_shutdown_script,shutdown_scriptpubkey,byte,shutdown_len",
    "msgtype,accept_channel2,65",
    "msgdata,accept_channel2,channel_id,channel_id,",
    "msgdata,accept_channel2,funding_satoshis,u64,",
    "msgdata,accept_channel2,dust_limit_satoshis,u64,",
    "msgdata,accept_channel2,max_htlc_value_in_flight_msat,u64,",
    "msgdata,accept_channel2,htlc_minimum_msat,u64,",
    "msgdata,accept_channel2,minimum_depth,u32,",
    "msgdata,accept_channel2,to_self_delay,u16,",
    "msgdata,accept_channel2,max_accepted_htlcs,u16,",
    "msgdata,accept_channel2,funding_pubkey,point,",
    "msgdata,accept_channel2,revocation_basepoint,point,",
    "msgdata,accept_channel2,payment_basepoint,point,",
    "msgdata,accept_channel2,delayed_payment_basepoint,point,",
    "msgdata,accept_channel2,htlc_basepoint,point,",
    "msgdata,accept_channel2,first_per_commitment_point,point,",
    "msgdata,accept_channel2,tlvs,accept_tlvs,",
    "tlvtype,accept_tlvs,option_upfront_shutdown_script,1",
    "tlvdata,accept_tlvs,option_upfront_shutdown_script,shutdown_len,u16,",
    "tlvdata,accept_tlvs,option_upfront_shutdown_script,shutdown_scriptpubkey,byte,shutdown_len",
    "msgtype,init_rbf,72",
    "msgdata,init_rbf,channel_id,channel_id,",
    "msgdata,init_rbf,funding_satoshis,u64,",
    "msgdata,init_rbf,locktime,u32,",
    "msgdata,init_rbf,funding_feerate_perkw,u32,",
    "msgtype,ack_rbf,73",
    "msgdata,ack_rbf,channel_id,channel_id,",
    "msgdata,ack_rbf,funding_satoshis,u64,",
]

# This is https://github.com/lightningnetwork/lightning-rfc/pull/880
channel_type_csv = [
    "tlvtype,open_channel_tlvs,channel_type,1",
    "tlvdata,open_channel_tlvs,channel_type,type,byte,...",
    "tlvtype,accept_channel_tlvs,channel_type,1",
    "tlvdata,accept_channel_tlvs,channel_type,type,byte,...",
]



#! /usr/bin/python3
from typing import Union, List


def bitfield_len(bitfield: Union[List[int], str]) -> int:
    """Return length of this field in bits (assuming it's a bitfield!)"""
    if isinstance(bitfield, str):
        return len(bytes.fromhex(bitfield)) * 8
    else:
        return len(bitfield) * 8


def has_bit(bitfield: Union[List[int], str], bitnum: int) -> bool:
    """Test bit in this bitfield (little-endian, as per BOLTs)"""
    bitlen = bitfield_len(bitfield)
    if bitnum >= bitlen:
        return False

    # internal to a msg, it's a list of int.
    if isinstance(bitfield, str):
        byte = bytes.fromhex(bitfield)[bitlen // 8 - 1 - bitnum // 8]
    else:
        byte = bitfield[bitlen // 8 - 1 - bitnum // 8]

    if (byte & (1 << (bitnum % 8))) != 0:
        return True
    else:
        return False


def bitfield(*args: int) -> str:
    """Create a bitfield hex value with these bit numbers set"""
    bytelen = (max(args) + 8) // 8
    bfield = bytearray(bytelen)
    for bitnum in args:
        bfield[bytelen - 1 - bitnum // 8] |= 1 << (bitnum % 8)
    return bfield.hex()



#! /usr/bin/python3
# This script exercises the backend implementation

# Released by Rusty Russell under CC0:
# https://creativecommons.org/publicdomain/zero/1.0/
from abc import ABC, abstractmethod


class Backend(ABC):
    """
    Generic implementation of Bitcoin backend
    This is useful when the LN node uses different type
    of bitcoin backend.
    """

    @abstractmethod
    def start(self) -> None:
        pass

    @abstractmethod
    def stop(self) -> None:
        pass

    @abstractmethod
    def restart(self) -> None:
        pass



# Support for funding txs.
import coincurve
import io
import logging

from typing import Tuple, Any, Optional, Union, Callable, Dict, List

from hashlib import sha256
from pyln.proto.message import Message

from .utils.bitcoin_utils import BitcoinUtils
from .utils import Side, privkey_expand
from .event import Event, ResolvableInt, ResolvableStr
from .namespace import namespace
from .runner import Runner
from .signature import Sig

import bitcoin.core.script as script
from bitcoin.core import (
    COutPoint,
    CScript,
    CTxIn,
    CTxOut,
    CMutableTransaction,
    CTxWitness,
    CTxInWitness,
    CScriptWitness,
    Hash160,
    CTransaction,
)
from bitcoin.wallet import P2WPKHBitcoinAddress

ResolvableFunding = Union["Funding", Callable[["Runner", "Event", str], "Funding"]]


def txid_raw(tx: str) -> str:
    """Helper to get the txid of a tx: note this is in wire protocol order, not bitcoin order!"""
    return CTransaction.deserialize(bytes.fromhex(tx)).GetTxid().hex()


class Funding(object):
    def __init__(
        self,
        funding_txid: str,
        funding_output_index: int,
        funding_amount: int,
        local_node_privkey: str,
        local_funding_privkey: str,
        remote_node_privkey: str,
        remote_funding_privkey: str,
        chain_hash: str = BitcoinUtils.blockchain_hash(),
        locktime: int = 0,
    ):
        self.chain_hash = chain_hash
        self.txid = funding_txid
        self.output_index = funding_output_index
        self.amount = funding_amount
        self.bitcoin_privkeys = [
            privkey_expand(local_funding_privkey),
            privkey_expand(remote_funding_privkey),
        ]
        self.node_privkeys = [
            privkey_expand(local_node_privkey),
            privkey_expand(remote_node_privkey),
        ]
        self.tx = None
        self.locktime = locktime
        self.outputs: List[Dict[str, Any]] = []
        self.inputs: List[Dict[str, Any]] = []

    def tx_hex(self) -> str:
        if not self.tx:
            return ""
        return self.tx.serialize().hex()

    @staticmethod
    def sort_by_keys(
        key_one: coincurve.PublicKey,
        key_two: coincurve.PublicKey,
        val_one: Any,
        val_two: Any,
    ) -> Tuple[Any, Any]:
        """In many places we have to sort elements into key or nodeid order"""
        # BOLT #3:
        # ## Funding Transaction Output
        #
        # * The funding output script is a P2WSH to: `2 <pubkey1> <pubkey2> 2
        #  OP_CHECKMULTISIG`
        # * Where `pubkey1` is the lexicographically lesser of the two
        #   `funding_pubkey` in compressed format, and where `pubkey2` is the
        #   lexicographically greater of the two.
        if key_one.format() < key_two.format():
            return val_one, val_two
        else:
            return val_two, val_one

    def node_id_sort(self, local: Any, remote: Any) -> Tuple[Any, Any]:
        """Sorts these two items into lexicographical node id order"""
        # BOLT #7:
        # - MUST set `node_id_1` and `node_id_2` to the public keys of the two
        #   nodes operating the channel, such that `node_id_1` is the
        #   lexicographically-lesser of the two compressed keys sorted in
        #   ascending lexicographic order.
        return self.sort_by_keys(
            self.node_id(Side.local), self.node_id(Side.remote), local, remote
        )

    @staticmethod
    def redeemscript_keys(
        key_one: coincurve.PublicKey, key_two: coincurve.PublicKey
    ) -> CScript:
        return CScript(
            [script.OP_2]
            + [
                k.format()
                for k in Funding.sort_by_keys(key_one, key_two, key_one, key_two)
            ]
            + [script.OP_2, script.OP_CHECKMULTISIG]
        )

    def redeemscript(self) -> CScript:
        key_a, key_b = self.funding_pubkeys_for_tx()
        return self.redeemscript_keys(key_a, key_b)

    @staticmethod
    def locking_script_keys(
        key_one: coincurve.PublicKey, key_two: coincurve.PublicKey
    ) -> CScript:
        return CScript(
            [script.OP_0, sha256(Funding.redeemscript_keys(key_one, key_two)).digest()]
        )

    def locking_script(self) -> CScript:
        a, b = self.funding_pubkeys_for_tx()
        return self.locking_script_keys(a, b)

    @staticmethod
    def start(
        local_node_privkey: str,
        local_funding_privkey: str,
        remote_node_privkey: str,
        remote_funding_privkey: str,
        funding_sats: int,
        locktime: int,
        chain_hash: str = BitcoinUtils.blockchain_hash(),
    ) -> "Funding":
        # Create dummy one to start: we will fill in txid at the end
        return Funding(
            "",
            0,
            funding_sats,
            local_node_privkey,
            local_funding_privkey,
            remote_node_privkey,
            remote_funding_privkey,
            chain_hash,
            locktime,
        )

    def add_input(
        self,
        serial_id: int,
        prevtx: str,
        prevtx_vout: int,
        script_sig: str,
        sequence: int,
        privkey: str = None,
    ) -> None:
        # the dummy runner sends empty info, skip
        if len(prevtx) == 0:
            return

        # Find the txid of the transaction
        prev_tx = CTransaction.deserialize(bytes.fromhex(prevtx))
        txin = CTxIn(COutPoint(prev_tx.GetTxid(), prevtx_vout), nSequence=sequence)

        # Get the previous output for its outscript + value
        prev_vout = prev_tx.vout[prevtx_vout]

        self.inputs.append(
            {
                "input": txin,
                "serial_id": serial_id,
                "sats": prev_vout.nValue,
                "prev_outscript": prev_vout.scriptPubKey.hex(),
                "redeemscript": script_sig,
                "privkey": privkey,
            }
        )

    def add_output(self, serial_id: int, script: str, sats: int) -> None:
        txout = CTxOut(sats, CScript(bytes.fromhex(script)))
        self.outputs.append({"output": txout, "serial_id": serial_id})

    def our_witnesses(self) -> str:
        """Extract expected witness data for our node"""
        witnesses = []
        # these were sorted in `build_tx`
        for _in in self.inputs:
            if not _in["privkey"]:
                continue

            wit = _in["sig"]
            print("witness is ... ", wit)
            elems = []
            for e in wit.scriptWitness.stack:
                elems.append("{{witness={0}}}".format(e.hex()))
            witnesses.append("{{witness_element=[{0}]}}".format(",".join(elems)))
        val = "[{}]".format(",".join(witnesses))
        print("witnesses are", val)
        return val

    def sign_our_inputs(self) -> None:
        assert self.tx is not None
        for idx, _in in enumerate(self.inputs):
            privkey = _in["privkey"]

            if privkey and "sig" not in _in:
                print("signing our input for tx", self.tx.serialize().hex())
                inkey = privkey_expand(privkey)
                inkey_pub = coincurve.PublicKey.from_secret(inkey.secret)

                # Really horrid hack to produce a signature for the
                # multisig utxo in tests/helpers.py
                if (
                    privkey
                    == "38204720bc4f9647fd58c6d0a4bd3a6dd2be16d8e4273c4d1bdd5774e8c51eaf"
                ):
                    redeemscript = bytes.fromhex(
                        "51210253cdf835e328346a4f19de099cf3d42d4a7041e073cd4057a1c4fd7cdbb1228f2103ae903722f21f85e651b8f9b18fc854084fb90eeb76452bdcfd0cb43a16a382a221036c264d68a9727afdc75949f7d7fa71910ae9ae8001a1fbffa6f7ce000976597c21036429fa8a4ef0b2b1d5cb553e34eeb90a32ab19fae1f0024f332ab4f74283a7282103d4232f19ea85051e7b76bf5f01d03e17eea8751463dee36d71413a739de1a92755ae"
                    )
                else:
                    address = P2WPKHBitcoinAddress.from_scriptPubKey(
                        CScript([script.OP_0, Hash160(inkey_pub.format())])
                    )
                    redeemscript = address.to_redeemScript()

                sighash = script.SignatureHash(
                    redeemscript,
                    self.tx,
                    idx,
                    script.SIGHASH_ALL,
                    amount=_in["sats"],
                    sigversion=script.SIGVERSION_WITNESS_V0,
                )
                sig = inkey.sign(sighash, hasher=None) + bytes([script.SIGHASH_ALL])

                if (
                    privkey
                    == "38204720bc4f9647fd58c6d0a4bd3a6dd2be16d8e4273c4d1bdd5774e8c51eaf"
                ):
                    _in["sig"] = CTxInWitness(
                        CScriptWitness([bytes([]), sig, redeemscript])
                    )
                else:
                    _in["sig"] = CTxInWitness(CScriptWitness([sig, inkey_pub.format()]))

    def add_witnesses(self, witness_stack: List[Dict[str, Any]]) -> str:
        assert self.tx is not None
        wits = []
        for idx, _in in enumerate(self.inputs):
            if "sig" in _in:
                wits.append(_in["sig"])
                continue

            if not len(witness_stack):
                continue

            elems = witness_stack.pop(0)["witness_element"]
            stack = []
            for elem in elems:
                stack.append(bytes.fromhex(elem["witness"]))

            wits.append(CTxInWitness(CScriptWitness(stack)))

        self.tx.wit = CTxWitness(wits)
        return self.tx.serialize().hex()

    def build_tx(self) -> str:
        # Sort inputs/outputs by serial number
        self.inputs = sorted(self.inputs, key=lambda k: k["serial_id"])
        self.outputs = sorted(self.outputs, key=lambda k: k["serial_id"])

        self.tx = CMutableTransaction(
            [i["input"] for i in self.inputs],
            [o["output"] for o in self.outputs],
            nVersion=2,
            nLockTime=self.locktime,
        )
        assert self.tx is not None
        self.txid = self.tx.GetTxid().hex()

        # Set the output index for the funding output
        locking_script = self.locking_script()
        for i, out in enumerate([o["output"] for o in self.outputs]):
            if out.scriptPubKey == locking_script:
                self.output_index = i
                self.amount = out.nValue

        return self.tx.serialize().hex()

    @staticmethod
    def from_utxo(
        txid_in: str,
        tx_index_in: int,
        sats: int,
        privkey: str,
        fee: int,
        local_node_privkey: str,
        local_funding_privkey: str,
        remote_node_privkey: str,
        remote_funding_privkey: str,
        chain_hash: str = BitcoinUtils.blockchain_hash(),
    ) -> Tuple["Funding", str]:
        """Make a funding transaction by spending this utxo using privkey: return Funding, tx."""

        # Create dummy one to start: we will fill in txid at the end.
        funding = Funding(
            "",
            0,
            sats - fee,
            local_node_privkey,
            local_funding_privkey,
            remote_node_privkey,
            remote_funding_privkey,
            chain_hash,
        )

        # input private key.
        inkey = privkey_expand(privkey)
        inkey_pub = coincurve.PublicKey.from_secret(inkey.secret)

        # use RBF'able input (requirement for dual-funded things)
        txin = CTxIn(
            COutPoint(bytes.fromhex(txid_in), tx_index_in), nSequence=0xFFFFFFFD
        )
        txout = CTxOut(
            sats - fee, CScript([script.OP_0, sha256(funding.redeemscript()).digest()])
        )
        tx = CMutableTransaction(
            [txin], [txout], nVersion=2, nLockTime=funding.locktime
        )

        # now fill in funding txid.
        funding.txid = tx.GetTxid().hex()
        funding.tx = tx

        # while we're here, sign the transaction.
        address = P2WPKHBitcoinAddress.from_scriptPubKey(
            CScript([script.OP_0, Hash160(inkey_pub.format())])
        )

        sighash = script.SignatureHash(
            address.to_redeemScript(),
            tx,
            0,
            script.SIGHASH_ALL,
            amount=sats,
            sigversion=script.SIGVERSION_WITNESS_V0,
        )
        sig = inkey.sign(sighash, hasher=None) + bytes([script.SIGHASH_ALL])

        tx.wit = CTxWitness([CTxInWitness(CScriptWitness([sig, inkey_pub.format()]))])
        return funding, tx.serialize().hex()

    def channel_id(self) -> str:
        # BOLT #2: This message introduces the `channel_id` to identify the
        # channel. It's derived from the funding transaction by combining the
        # `funding_txid` and the `funding_output_index`, using big-endian
        # exclusive-OR (i.e. `funding_output_index` alters the last 2 bytes).
        chanid = bytearray.fromhex(self.txid)
        chanid[-1] ^= self.output_index % 256
        chanid[-2] ^= self.output_index // 256
        return chanid.hex()

    @staticmethod
    def funding_pubkey_key(privkey: coincurve.PrivateKey) -> coincurve.PublicKey:
        return coincurve.PublicKey.from_secret(privkey.secret)

    def funding_pubkey(self, side: Side) -> coincurve.PublicKey:
        return self.funding_pubkey_key(self.bitcoin_privkeys[side])

    def funding_pubkeys_for_tx(self) -> Tuple[coincurve.PublicKey, coincurve.PublicKey]:
        """Returns funding pubkeys, in tx order"""
        # BOLT #3:
        # ## Funding Transaction Output
        #
        # * The funding output script is a P2WSH to: `2 <pubkey1> <pubkey2> 2
        #  OP_CHECKMULTISIG`
        # * Where `pubkey1` is the lexicographically lesser of the two
        #   `funding_pubkey` in compressed format, and where `pubkey2` is the
        #   lexicographically greater of the two.
        return self.sort_by_keys(
            self.funding_pubkey(Side.local),
            self.funding_pubkey(Side.remote),
            self.funding_pubkey(Side.local),
            self.funding_pubkey(Side.remote),
        )

    def funding_privkeys_for_tx(
        self,
    ) -> Tuple[coincurve.PrivateKey, coincurve.PrivateKey]:
        """Returns funding private keys, in tx order"""
        return self.sort_by_keys(
            self.funding_pubkey(Side.local),
            self.funding_pubkey(Side.remote),
            self.bitcoin_privkeys[Side.local],
            self.bitcoin_privkeys[Side.remote],
        )

    def node_id(self, side: Side) -> coincurve.PublicKey:
        return coincurve.PublicKey.from_secret(self.node_privkeys[side].secret)

    def node_ids(self) -> Tuple[coincurve.PublicKey, coincurve.PublicKey]:
        """Returns node pubkeys, in order"""
        return self.node_id_sort(self.node_id(Side.local), self.node_id(Side.remote))

    def node_id_privkeys(self) -> Tuple[coincurve.PrivateKey, coincurve.PrivateKey]:
        """Returns node private keys, in order"""
        return self.node_id_sort(
            self.node_privkeys[Side.local], self.node_privkeys[Side.remote]
        )

    def funding_pubkeys_for_gossip(
        self,
    ) -> Tuple[coincurve.PublicKey, coincurve.PublicKey]:
        """Returns funding public keys, in gossip order"""
        return self.node_id_sort(
            self.funding_pubkey(Side.local), self.funding_pubkey(Side.remote)
        )

    def funding_privkeys_for_gossip(
        self,
    ) -> Tuple[coincurve.PublicKey, coincurve.PublicKey]:
        """Returns funding private keys, in gossip order"""
        return self.node_id_sort(
            self.bitcoin_privkeys[Side.local], self.bitcoin_privkeys[Side.remote]
        )

    def _unsigned_channel_announcment(
        self, features: str, short_channel_id: str
    ) -> Message:
        """Produce a channel_announcement message with dummy sigs"""
        node_ids = self.node_ids()
        bitcoin_keys = self.funding_pubkeys_for_gossip()
        return Message(
            namespace().get_msgtype("channel_announcement"),
            node_signature_1=Sig(bytes(64)),
            node_signature_2=Sig(bytes(64)),
            bitcoin_signature_1=Sig(bytes(64)),
            bitcoin_signature_2=Sig(bytes(64)),
            features=features,
            chain_hash=self.chain_hash,
            short_channel_id=short_channel_id,
            node_id_1=node_ids[0].format(),
            node_id_2=node_ids[1].format(),
            bitcoin_key_1=bitcoin_keys[0].format(),
            bitcoin_key_2=bitcoin_keys[1].format(),
        )

    def channel_announcement(self, short_channel_id: str, features: str) -> Message:
        """Produce a (signed) channel_announcement message"""
        ann = self._unsigned_channel_announcment(features, short_channel_id)
        # BOLT #7:
        # - MUST compute the double-SHA256 hash `h` of the message, beginning
        #   at offset 256, up to the end of the message.
        #   - Note: the hash skips the 4 signatures but hashes the rest of the
        #     message, including any future fields appended to the end.
        buf = io.BytesIO()
        ann.write(buf)
        # Note the first two 'type' bytes!
        h = sha256(sha256(buf.getvalue()[2 + 256 :]).digest()).digest()

        # BOLT #7:
        # - MUST set `node_signature_1` and `node_signature_2` to valid
        #   signatures of the hash `h` (using `node_id_1` and `node_id_2`'s
        #   respective secrets).
        node_privkeys = self.node_id_privkeys()
        ann.set_field("node_signature_1", Sig(node_privkeys[0].secret.hex(), h.hex()))
        ann.set_field("node_signature_2", Sig(node_privkeys[1].secret.hex(), h.hex()))

        bitcoin_privkeys = self.funding_privkeys_for_gossip()
        # - MUST set `bitcoin_signature_1` and `bitcoin_signature_2` to valid
        #   signatures of the hash `h` (using `bitcoin_key_1` and
        #   `bitcoin_key_2`'s respective secrets).
        ann.set_field(
            "bitcoin_signature_1", Sig(bitcoin_privkeys[0].secret.hex(), h.hex())
        )
        ann.set_field(
            "bitcoin_signature_2", Sig(bitcoin_privkeys[1].secret.hex(), h.hex())
        )

        return ann

    def channel_update(
        self,
        short_channel_id: str,
        side: Side,
        disable: bool,
        cltv_expiry_delta: int,
        htlc_minimum_msat: int,
        fee_base_msat: int,
        fee_proportional_millionths: int,
        timestamp: int,
        htlc_maximum_msat: Optional[int],
    ) -> Message:
        # BOLT #7: The `channel_flags` bitfield is used to indicate the
        # direction of the channel: it identifies the node that this update
        # originated from and signals various options concerning the
        # channel. The following table specifies the meaning of its individual
        # bits:
        #
        # | Bit Position  | Name        | Meaning                          |
        # | ------------- | ----------- | -------------------------------- |
        # | 0             | `direction` | Direction this update refers to. |
        # | 1             | `disable`   | Disable the channel.             |

        # BOLT #7:
        #   - if the origin node is `node_id_1` in the message:
        #     - MUST set the `direction` bit of `channel_flags` to 0.
        #   - otherwise:
        #     - MUST set the `direction` bit of `channel_flags` to 1.
        if self.funding_pubkey(side) == self.funding_pubkeys_for_gossip()[0]:
            channel_flags = 0
        else:
            channel_flags = 1

        if disable:
            channel_flags |= 2

        # BOLT #7: The message_flags bitfield is used to provide additional
        # details about the message:
        #
        # | Bit Position  | Name           |
        # | ------------- | ---------------|
        # | 0             | `must_be_one`  |
        # | 1             | `dont_forward` |
        message_flags = 1

        # Begin with a fake signature.
        update = Message(
            namespace().get_msgtype("channel_update"),
            short_channel_id=short_channel_id,
            signature=Sig(bytes(64)),
            chain_hash=self.chain_hash,
            timestamp=timestamp,
            message_flags=message_flags,
            channel_flags=channel_flags,
            cltv_expiry_delta=cltv_expiry_delta,
            htlc_minimum_msat=htlc_minimum_msat,
            fee_base_msat=fee_base_msat,
            fee_proportional_millionths=fee_proportional_millionths,
        )
        if htlc_maximum_msat:
            update.set_field("htlc_maximum_msat", htlc_maximum_msat)

        # BOLT #7:
        # - MUST set `signature` to the signature of the double-SHA256 of the
        #   entire remaining packet after `signature`, using its own `node_id`.
        buf = io.BytesIO()
        update.write(buf)
        # Note the first two 'type' bytes!
        h = sha256(sha256(buf.getvalue()[2 + 64 :]).digest()).digest()

        update.set_field(
            "signature", Sig(self.node_privkeys[side].secret.hex(), h.hex())
        )

        return update

    def node_announcement(
        self,
        side: Side,
        features: str,
        rgb_color: Tuple[int, int, int],
        alias: str,
        addresses: bytes,
        timestamp: int,
    ) -> Message:
        # Begin with a fake signature.
        ann = Message(
            namespace().get_msgtype("node_announcement"),
            signature=Sig(bytes(64)),
            features=features,
            timestamp=timestamp,
            node_id=self.node_id(side).format().hex(),
            rgb_color=bytes(rgb_color).hex(),
            alias=bytes(alias, encoding="utf-8").zfill(32),
            addresses=addresses,
        )

        # BOLT #7:
        #  - MUST set `signature` to the signature of the double-SHA256 of the entire
        #  remaining packet after `signature` (using the key given by `node_id`).
        buf = io.BytesIO()
        ann.write(buf)
        # Note the first two 'type' bytes!
        h = sha256(sha256(buf.getvalue()[2 + 64 :]).digest()).digest()

        ann.set_field("signature", Sig(self.node_privkeys[side].secret.hex(), h.hex()))
        return ann

    def close_tx(self, fee: int, privkey_dest: str) -> str:
        """Create a (mutual) close tx"""
        txin = CTxIn(COutPoint(bytes.fromhex(self.txid), self.output_index))

        out_privkey = privkey_expand(privkey_dest)

        txout = CTxOut(
            self.amount - fee,
            CScript(
                [
                    script.OP_0,
                    Hash160(
                        coincurve.PublicKey.from_secret(out_privkey.secret).format()
                    ),
                ]
            ),
        )

        tx = CMutableTransaction(vin=[txin], vout=[txout])
        sighash = script.SignatureHash(
            self.redeemscript(),
            tx,
            inIdx=0,
            hashtype=script.SIGHASH_ALL,
            amount=self.amount,
            sigversion=script.SIGVERSION_WITNESS_V0,
        )

        sigs = [
            key.sign(sighash, hasher=None) for key in self.funding_privkeys_for_tx()
        ]
        # BOLT #3:
        # ## Closing Transaction
        # ...
        #    * `txin[0]` witness: `0 <signature_for_pubkey1> <signature_for_pubkey2>`
        witness = CScriptWitness(
            [
                bytes(),
                sigs[0] + bytes([script.SIGHASH_ALL]),
                sigs[1] + bytes([script.SIGHASH_ALL]),
                self.redeemscript(),
            ]
        )
        tx.wit = CTxWitness([CTxInWitness(witness)])
        return tx.serialize().hex()


class AcceptFunding(Event):
    """Event to accept funding information from a peer.  Stashes 'Funding'."""

    def __init__(
        self,
        funding_txid: ResolvableStr,
        funding_output_index: ResolvableInt,
        funding_amount: ResolvableInt,
        local_node_privkey: ResolvableStr,
        local_funding_privkey: ResolvableStr,
        remote_node_privkey: ResolvableStr,
        remote_funding_privkey: ResolvableStr,
        chain_hash: str = BitcoinUtils.blockchain_hash(),
    ):
        super().__init__()
        self.funding_txid = funding_txid
        self.funding_output_index = funding_output_index
        self.funding_amount = funding_amount
        self.local_node_privkey = local_node_privkey
        self.local_funding_privkey = local_funding_privkey
        self.remote_node_privkey = remote_node_privkey
        self.remote_funding_privkey = remote_funding_privkey
        self.chain_hash = chain_hash

    def action(self, runner: Runner) -> bool:
        super().action(runner)

        funding = Funding(
            chain_hash=self.chain_hash,
            **self.resolve_args(
                runner,
                {
                    "funding_txid": self.funding_txid,
                    "funding_output_index": self.funding_output_index,
                    "funding_amount": self.funding_amount,
                    "local_node_privkey": self.local_node_privkey,
                    "local_funding_privkey": self.local_funding_privkey,
                    "remote_node_privkey": self.remote_node_privkey,
                    "remote_funding_privkey": self.remote_funding_privkey,
                },
            ),
        )
        runner.add_stash("Funding", funding)
        return True


class CreateFunding(Event):
    """Event to create a funding tx from a P2WPKH UTXO.  Stashes 'Funding' and 'FundingTx'."""

    def __init__(
        self,
        txid_in: str,
        tx_index_in: int,
        sats_in: int,
        spending_privkey: str,
        fee: int,
        local_node_privkey: ResolvableStr,
        local_funding_privkey: ResolvableStr,
        remote_node_privkey: ResolvableStr,
        remote_funding_privkey: ResolvableStr,
        chain_hash: str = BitcoinUtils.blockchain_hash(),
    ):
        super().__init__()
        self.txid_in = txid_in
        self.tx_index_in = tx_index_in
        self.sats_in = sats_in
        self.spending_privkey = spending_privkey
        self.fee = fee
        self.local_node_privkey = local_node_privkey
        self.local_funding_privkey = local_funding_privkey
        self.remote_node_privkey = remote_node_privkey
        self.remote_funding_privkey = remote_funding_privkey
        self.chain_hash = chain_hash

    def action(self, runner: Runner) -> bool:
        super().action(runner)

        funding, tx = Funding.from_utxo(
            self.txid_in,
            self.tx_index_in,
            self.sats_in,
            self.spending_privkey,
            self.fee,
            chain_hash=self.chain_hash,
            **self.resolve_args(
                runner,
                {
                    "local_node_privkey": self.local_node_privkey,
                    "local_funding_privkey": self.local_funding_privkey,
                    "remote_node_privkey": self.remote_node_privkey,
                    "remote_funding_privkey": self.remote_funding_privkey,
                },
            ),
        )

        runner.add_stash("Funding", funding)
        runner.add_stash("FundingTx", tx)
        return True


class CreateDualFunding(Event):
    """Event to create a 'dual-funded' funding tx. Stashes 'Funding'"""

    def __init__(
        self,
        fee: int,
        funding_sats: ResolvableInt,
        locktime: ResolvableInt,
        local_node_privkey: str,
        local_funding_privkey: str,
        remote_node_privkey: str,
        remote_funding_privkey: ResolvableStr,
        chain_hash: str = BitcoinUtils.blockchain_hash(),
    ):
        super().__init__()

        self.funding_sats = funding_sats
        self.locktime = locktime
        self.local_node_privkey = local_node_privkey
        self.local_funding_privkey = local_funding_privkey
        self.remote_node_privkey = remote_node_privkey
        self.remote_funding_privkey = remote_funding_privkey
        self.chain_hash = chain_hash

    def action(self, runner: Runner) -> bool:
        super().action(runner)

        funding = Funding.start(
            local_node_privkey=self.local_node_privkey,
            local_funding_privkey=self.local_funding_privkey,
            chain_hash=self.chain_hash,
            **self.resolve_args(
                runner,
                {
                    "funding_sats": self.funding_sats,
                    "remote_funding_privkey": self.remote_funding_privkey,
                    "remote_node_privkey": self.remote_node_privkey,
                    "locktime": self.locktime,
                },
            ),
        )

        runner.add_stash("Funding", funding)

        return True


class AddInput(Event):
    def __init__(
        self,
        funding: ResolvableFunding,
        serial_id: ResolvableInt,
        prevtx: ResolvableStr,
        prevtx_vout: ResolvableInt,
        script_sig: ResolvableStr,
        sequence: ResolvableInt = 0xFFFFFFFD,
        privkey: str = None,
    ):
        super().__init__()
        self.funding = funding
        self.privkey = privkey
        self.serial_id = serial_id
        self.prevtx = prevtx
        self.prevtx_vout = prevtx_vout
        self.script_sig = script_sig
        self.sequence = sequence

    def action(self, runner: Runner) -> bool:
        super().action(runner)
        funding = self.resolve_arg("funding", runner, self.funding)
        funding.add_input(
            **self.resolve_args(
                runner,
                {
                    "serial_id": self.serial_id,
                    "prevtx": self.prevtx,
                    "prevtx_vout": self.prevtx_vout,
                    "script_sig": self.script_sig,
                    "sequence": self.sequence,
                    "privkey": self.privkey,
                },
            )
        )
        return True


class AddOutput(Event):
    def __init__(
        self,
        funding: ResolvableFunding,
        serial_id: ResolvableInt,
        sats: ResolvableInt,
        script: ResolvableStr,
    ):
        super().__init__()
        self.funding = funding
        self.serial_id = serial_id
        self.sats = sats
        self.script = script

    def action(self, runner: Runner) -> bool:
        super().action(runner)
        funding = self.resolve_arg("funding", runner, self.funding)
        funding.add_output(
            **self.resolve_args(
                runner,
                {"serial_id": self.serial_id, "sats": self.sats, "script": self.script},
            )
        )
        return True


class FinalizeFunding(Event):
    def __init__(self, funding: ResolvableFunding):
        self.funding = funding

    def action(self, runner: Runner) -> bool:
        funding = self.resolve_arg("funding", runner, self.funding)

        tx = funding.build_tx()
        funding.sign_our_inputs()
        # FIXME: sanity checks?
        logging.debug(f"finalized funding {tx}")
        return True


class AddWitnesses(Event):
    def __init__(
        self,
        funding: ResolvableFunding,
        witness_stack: Union[
            List[Dict[str, Any]],
            Callable[["Runner", "Event", str], List[Dict[str, Any]]],
        ],
    ):
        self.funding = funding
        self.witness_stack = witness_stack

    def action(self, runner: Runner) -> bool:
        funding = self.resolve_arg("funding", runner, self.funding)
        stack = self.resolve_arg("witness_stack", runner, self.witness_stack)
        # FIXME: is there a way to resolve this more .. nicely?
        # Convert from string to python obj
        wit_stack = eval(stack)
        tx_hex = funding.add_witnesses(wit_stack)
        runner.add_stash("FundingTx", tx_hex)
        return True



#!/usr/bin/python3
# This script exercises the core-lightning implementation

# Released by Rusty Russell under CC0:
# https://creativecommons.org/publicdomain/zero/1.0/

import hashlib
import pyln.client
import pyln.proto.wire
import os
import subprocess
import lnprototest
import bitcoin.core
import struct
import shutil
import logging
import socket
import time

from contextlib import closing
from datetime import date
from concurrent import futures
from lnprototest.backend import Bitcoind
from lnprototest import (
    Event,
    EventError,
    SpecFileError,
    KeySet,
    Conn,
    namespace,
    MustNotMsg,
)
from lnprototest import wait_for
from typing import Dict, Any, Callable, List, Optional, cast

TIMEOUT = int(os.getenv("TIMEOUT", "60"))
LIGHTNING_SRC = os.path.join(os.getcwd(), os.getenv("LIGHTNING_SRC", "../lightning/"))


class CLightningConn(lnprototest.Conn):
    def __init__(self, connprivkey: str, port: int):
        super().__init__(connprivkey)
        # FIXME: pyln.proto.wire should just use coincurve PrivateKey!
        self.connection = pyln.proto.wire.connect(
            pyln.proto.wire.PrivateKey(bytes.fromhex(self.connprivkey.to_hex())),
            # FIXME: Ask node for pubkey
            pyln.proto.wire.PublicKey(
                bytes.fromhex(
                    "0279be667ef9dcbbac55a06295ce870b07029bfcdb2dce28d959f2815b16f81798"
                )
            ),
            "127.0.0.1",
            port,
        )


class Runner(lnprototest.Runner):
    def __init__(self, config: Any):
        super().__init__(config)
        self.running = False
        self.rpc = None
        self.bitcoind = None
        self.proc = None
        self.cleanup_callbacks: List[Callable[[], None]] = []
        self.fundchannel_future: Optional[Any] = None
        self.is_fundchannel_kill = False
        self.executor = futures.ThreadPoolExecutor(max_workers=20)

        self.startup_flags = []
        for flag in config.getoption("runner_args"):
            self.startup_flags.append("--{}".format(flag))

        opts = (
            subprocess.run(
                [
                    "{}/lightningd/lightningd".format(LIGHTNING_SRC),
                    "--list-features-only",
                ],
                stdout=subprocess.PIPE,
                check=True,
            )
            .stdout.decode("utf-8")
            .splitlines()
        )
        self.options: Dict[str, str] = {}
        for o in opts:
            if o.startswith("supports_"):
                self.options[o] = "true"
            else:
                k, v = o.split("/")
                self.options[k] = v

    def __reserve(self) -> int:
        """
        When python asks for a free port from the os, it is possible that
        with concurrent access, the port that is picked is a port that is not free
        anymore when we go to bind the daemon like bitcoind port.

        Source: https://stackoverflow.com/questions/1365265/on-localhost-how-do-i-pick-a-free-port-number
        """
        with closing(socket.socket(socket.AF_INET, socket.SOCK_STREAM)) as s:
            s.bind(("", 0))
            s.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
            return s.getsockname()[1]

    def __init_sandbox_dir(self) -> None:
        """Create the tmp directory for lnprotest and lightningd"""
        self.lightning_dir = os.path.join(self.directory, "lightningd")
        if not os.path.exists(self.lightning_dir):
            os.makedirs(self.lightning_dir)

    def get_keyset(self) -> KeySet:
        return KeySet(
            revocation_base_secret="0000000000000000000000000000000000000000000000000000000000000011",
            payment_base_secret="0000000000000000000000000000000000000000000000000000000000000012",
            delayed_payment_base_secret="0000000000000000000000000000000000000000000000000000000000000013",
            htlc_base_secret="0000000000000000000000000000000000000000000000000000000000000014",
            shachain_seed="FF" * 32,
        )

    def get_node_privkey(self) -> str:
        return "01"

    def get_node_bitcoinkey(self) -> str:
        return "0000000000000000000000000000000000000000000000000000000000000010"

    def is_running(self) -> bool:
        return self.running

    def start(self, also_bitcoind: bool = True) -> None:
        self.logger.debug("[START]")
        self.__init_sandbox_dir()
        self.lightning_port = self.__reserve()
        if also_bitcoind:
            self.bitcoind = Bitcoind(self.directory)
            try:
                self.bitcoind.start()
            except Exception as ex:
                self.logger.debug(f"Exception with message {ex}")
            self.logger.debug("RUN Bitcoind")
        self.proc = subprocess.Popen(
            [
                "{}/lightningd/lightningd".format(LIGHTNING_SRC),
                "--lightning-dir={}".format(self.lightning_dir),
                "--funding-confirms=3",
                "--dev-force-privkey=0000000000000000000000000000000000000000000000000000000000000001",
                "--dev-force-bip32-seed=0000000000000000000000000000000000000000000000000000000000000001",
                "--dev-force-channel-secrets=0000000000000000000000000000000000000000000000000000000000000010/0000000000000000000000000000000000000000000000000000000000000011/0000000000000000000000000000000000000000000000000000000000000012/0000000000000000000000000000000000000000000000000000000000000013/0000000000000000000000000000000000000000000000000000000000000014/FFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFF",
                "--dev-bitcoind-poll=1",
                "--dev-fast-gossip",
                "--dev-allow-localhost",
                "--dev-no-htlc-timeout",
                "--bind-addr=127.0.0.1:{}".format(self.lightning_port),
                "--network=regtest",
                "--bitcoin-rpcuser=rpcuser",
                "--bitcoin-rpcpassword=rpcpass",
                f"--bitcoin-rpcconnect=localhost:{self.bitcoind.port}",
                "--log-level=io",
                "--log-file=log",
                "--htlc-maximum-msat=2000sat",
            ]
            + self.startup_flags
        )
        self.running = True
        self.rpc = pyln.client.LightningRpc(
            os.path.join(self.lightning_dir, "regtest", "lightning-rpc")
        )
        self.logger.debug("RUN core-lightning")

        def node_ready(rpc: pyln.client.LightningRpc) -> bool:
            try:
                rpc.getinfo()
                return True
            except Exception as ex:
                logging.debug(f"waiting for core-lightning: Exception received {ex}")
                return False

        wait_for(lambda: node_ready(self.rpc), timeout=TIMEOUT)
        logging.debug("Waited for core-lightning")

        # Make sure that we see any funds that come to our wallet
        for i in range(5):
            self.rpc.newaddr()

    def shutdown(self, also_bitcoind: bool = True) -> None:
        for cb in self.cleanup_callbacks:
            cb()
        self.rpc.stop()
        if also_bitcoind:
            self.bitcoind.stop()

    def stop(self, print_logs: bool = False, also_bitcoind: bool = True) -> None:
        self.logger.debug("[STOP]")
        self.shutdown(also_bitcoind=also_bitcoind)
        self.running = False
        for c in self.conns.values():
            cast(CLightningConn, c).connection.connection.close()
        if print_logs:
            log_path = f"{self.lightning_dir}/regtest/log"
            with open(log_path) as log:
                self.logger.info("---------- core-lightning logging ----------------")
                self.logger.info(log.read())
                # now we make a backup of the log
                shutil.copy(
                    log_path,
                    f'/tmp/c-lightning-log_{date.today().strftime("%b-%d-%Y_%H:%M:%S")}',
                )
        shutil.rmtree(os.path.join(self.lightning_dir, "regtest"))

    def restart(self) -> None:
        self.logger.debug("[RESTART]")
        self.stop(also_bitcoind=False)
        # Make a clean start
        super().restart()
        self.bitcoind.restart()
        self.start(also_bitcoind=False)

    def connect(self, _: Event, connprivkey: str) -> None:
        self.add_conn(CLightningConn(connprivkey, self.lightning_port))

    def getblockheight(self) -> int:
        return self.bitcoind.rpc.getblockcount()

    def trim_blocks(self, newheight: int) -> None:
        h = self.bitcoind.rpc.getblockhash(newheight + 1)
        self.bitcoind.rpc.invalidateblock(h)

    def add_blocks(self, event: Event, txs: List[str], n: int) -> None:
        for tx in txs:
            self.bitcoind.rpc.sendrawtransaction(tx)
        self.bitcoind.rpc.generatetoaddress(n, self.bitcoind.rpc.getnewaddress())

        wait_for(lambda: self.rpc.getinfo()["blockheight"] == self.getblockheight())

    def recv(self, event: Event, conn: Conn, outbuf: bytes) -> None:
        try:
            cast(CLightningConn, conn).connection.send_message(outbuf)
        except BrokenPipeError:
            # This happens when they've sent an error and closed; try
            # reading it to figure out what went wrong.
            fut = self.executor.submit(
                cast(CLightningConn, conn).connection.read_message
            )
            try:
                msg = fut.result(1)
            except futures.TimeoutError:
                msg = None
            if msg:
                raise EventError(
                    event, "Connection closed after sending {}".format(msg.hex())
                )
            else:
                raise EventError(event, "Connection closed")

    def fundchannel(
        self,
        event: Event,
        conn: Conn,
        amount: int,
        feerate: int = 253,
        expect_fail: bool = False,
    ) -> None:
        """
        event       - the event which cause this, for error logging
        conn        - which conn (i.e. peer) to fund.
        amount      - amount to fund the channel with
        feerate     - feerate, in kiloweights
        expect_fail - true if this command is expected to error/fail
        """
        # First, check that another fundchannel isn't already running
        if self.fundchannel_future:
            if not self.fundchannel_future.done():
                raise RuntimeError(
                    "{} called fundchannel while another channel funding (fundchannel/init_rbf) is still in process".format(
                        event
                    )
                )
            self.fundchannel_future = None

        def _fundchannel(
            runner: Runner,
            conn: Conn,
            amount: int,
            feerate: int,
            expect_fail: bool = False,
        ) -> str:
            peer_id = conn.pubkey.format().hex()
            # Need to supply feerate here, since regtest cannot estimate fees
            try:
                return (
                    runner.rpc.fundchannel(
                        peer_id, amount, feerate="{}perkw".format(feerate)
                    ),
                    False,
                )
            except Exception as ex:
                # FIXME: this should not return None
                # but for now that we do not have any
                # use case where returni value is needed
                # we keep return null.
                #
                # The main reason to do this mess
                # is that in lnprototest do not have
                # any custom way to report a spec violation
                # failure, so for this reason we have different exception
                # at the same time (because this mess is needed to make stuff async
                # and look at exchanged message before finish the call). So
                # the solution is that we log the RPC exception (this may cause a spec
                # validation failure) and we care just the lnprototest exception as
                # real reason to abort.
                logging.error(f"{ex}")
                return str(ex), True

        def _done(fut: Any) -> None:
            result, ok = fut.result()
            if not ok and not self.is_fundchannel_kill and not expect_fail:
                raise Exception(result)
            self.fundchannel_future = None
            self.is_fundchannel_kill = False
            self.cleanup_callbacks.remove(self.kill_fundchannel)

        # FIXME core lightning has a race condition
        # when the core lightning node will go to fund the channel
        # but it will go to to connect with the node before
        # This required some more analysis from core lightning side
        time.sleep(1)

        fut = self.executor.submit(
            _fundchannel, self, conn, amount, feerate, expect_fail
        )
        fut.add_done_callback(_done)
        self.fundchannel_future = fut
        self.cleanup_callbacks.append(self.kill_fundchannel)

    def close_channel(self, channel_id: str) -> None:
        logging.debug("[CLOSE CHANNEL with channel id: '{}']".format(channel_id))
        self.rpc.close(peer_id=channel_id)

    def kill_fundchannel(self) -> None:
        fut = self.fundchannel_future
        self.fundchannel_future = None
        self.is_fundchannel_kill = True

        if fut:
            try:
                fut.result(0)
            except (SpecFileError, futures.TimeoutError):
                pass
            except Exception as ex:
                raise ex from None

    def init_rbf(
        self,
        event: Event,
        conn: Conn,
        channel_id: str,
        amount: int,
        utxo_txid: str,
        utxo_outnum: int,
        feerate: int,
    ) -> None:
        if self.fundchannel_future:
            self.kill_fundchannel()

        startweight = 42 + 172  # base weight, funding output
        # Build a utxo using the given utxo
        fmt_feerate = "{}perkw".format(feerate)
        utxos = ["{}:{}".format(utxo_txid, utxo_outnum)]
        initial_psbt = self.rpc.utxopsbt(
            amount,
            fmt_feerate,
            startweight,
            utxos,
            reservedok=True,
            min_witness_weight=110,
            locktime=0,
            excess_as_change=True,
        )["psbt"]

        def _run_rbf(runner: Runner, conn: Conn) -> Dict[str, Any]:
            bump = runner.rpc.openchannel_bump(
                channel_id, amount, initial_psbt, funding_feerate=fmt_feerate
            )
            update = runner.rpc.openchannel_update(channel_id, bump["psbt"])

            # Run until they're done sending us updates
            while not update["commitments_secured"]:
                update = runner.rpc.openchannel_update(channel_id, update["psbt"])
            signed_psbt = runner.rpc.signpsbt(update["psbt"])["signed_psbt"]
            return runner.rpc.openchannel_signed(channel_id, signed_psbt)

        def _done(fut: Any) -> None:
            exception = fut.exception(0)
            if exception:
                raise (exception)

        fut = self.executor.submit(_run_rbf, self, conn)
        fut.add_done_callback(_done)

    def invoice(self, event: Event, amount: int, preimage: str) -> None:
        self.rpc.invoice(
            msatoshi=amount,
            label=str(event),
            description="invoice from {}".format(event),
            preimage=preimage,
        )

    def accept_add_fund(self, event: Event) -> None:
        self.rpc.call(
            "funderupdate",
            {
                "policy": "match",
                "policy_mod": 100,
                "fuzz_percent": 0,
                "leases_only": False,
            },
        )

    def addhtlc(self, event: Event, conn: Conn, amount: int, preimage: str) -> None:
        payhash = hashlib.sha256(bytes.fromhex(preimage)).hexdigest()
        routestep = {
            "msatoshi": amount,
            "id": conn.pubkey.format().hex(),
            # We internally add one.
            "delay": 4,
            # We actually ignore this.
            "channel": "1x1x1",
        }
        self.rpc.sendpay([routestep], payhash)

    def get_output_message(
        self, conn: Conn, event: Event, timeout: int = TIMEOUT
    ) -> Optional[bytes]:
        fut = self.executor.submit(cast(CLightningConn, conn).connection.read_message)
        try:
            return fut.result(timeout)
        except futures.TimeoutError as ex:
            logging.error(f"timeout exception {ex}")
            return None
        except Exception as ex:
            logging.error(f"{ex}")
            return None

    def check_error(self, event: Event, conn: Conn) -> Optional[str]:
        # We get errors in form of err msgs, always.
        super().check_error(event, conn)
        msg = self.get_output_message(conn, event)
        if msg is None:
            return None
        return msg.hex()

    def check_final_error(
        self,
        event: Event,
        conn: Conn,
        expected: bool,
        must_not_events: List[MustNotMsg],
    ) -> None:
        if not expected:
            # Inject raw packet to ensure it hangs up *after* processing all previous ones.
            cast(CLightningConn, conn).connection.connection.send(bytes(18))

            while True:
                binmsg = self.get_output_message(conn, event)
                if binmsg is None:
                    break
                for e in must_not_events:
                    if e.matches(binmsg):
                        raise EventError(
                            event, "Got msg banned by {}: {}".format(e, binmsg.hex())
                        )

                # Don't assume it's a message type we know!
                msgtype = struct.unpack(">H", binmsg[:2])[0]
                if msgtype == namespace().get_msgtype("error").number:
                    raise EventError(event, "Got error msg: {}".format(binmsg.hex()))

        cast(CLightningConn, conn).connection.connection.close()

    def expect_tx(self, event: Event, txid: str) -> None:
        # Ah bitcoin endianness...
        revtxid = bitcoin.core.lx(txid).hex()

        # This txid should appear in the mempool.
        try:
            wait_for(lambda: revtxid in self.bitcoind.rpc.getrawmempool())
        except ValueError:
            raise EventError(
                event,
                "Did not broadcast the txid {}, just {}".format(
                    revtxid,
                    [
                        (txid, self.bitcoind.rpc.getrawtransaction(txid))
                        for txid in self.bitcoind.rpc.getrawmempool()
                    ],
                ),
            )

    def has_option(self, optname: str) -> Optional[str]:
        """Returns None if it doesn't support, otherwise 'even' or 'odd' (required or supported)"""
        if optname in self.options:
            return self.options[optname]
        return None

    def add_startup_flag(self, flag: str) -> None:
        logging.debug("[ADD STARTUP FLAG '{}']".format(flag))
        self.startup_flags.append("--{}".format(flag))



"""
Lightning network Speck utils, is a collection of methods that helps to
work with some concept of lightning network RFC

It also contains a method to generate the correct sequence of channel opening
and, and it feels a dictionary with all the propriety that needs to
be used after this sequence of steps.

author: Vincenzo PAlazzo https://github.com/vincenzopalazzo
"""
from typing import List, Optional


class LightningUtils:
    """
    Main implementation class of the lightning networks utils.

    The implementation class contains only static methods that
    apply the rules specified in the lightning network RFC.
    """

    @staticmethod
    def derive_short_channel_id(block_height: int, tx_idx: int, tx_output) -> str:
        """
        Derive the short channel id with the specified
        parameters, and return the result as string.

        RFC definition: https://github.com/lightning/bolts/blob/93909f67f6a48ee3f155a6224c182e612dd5f187/07-routing-gossip.md#definition-of-short_channel_id

        The short_channel_id is the unique description of the funding transaction. It is constructed as follows:
            - the most significant 3 bytes: indicating the block height
            - the next 3 bytes: indicating the transaction index within the block
            - the least significant 2 bytes: indicating the output index that pays to the channel.

        e.g: a short_channel_id might be written as 539268x845x1, indicating a channel on the
        output 1 of the transaction at index 845 of the block at height 539268.

        block_height: str
            Block height.
        tx_idx: int
            Transaction index inside the block.
        tx_output: int
            Output index inside the transaction.
        """
        return f"{block_height}x{tx_idx}x{tx_output}"


def connect_to_node_helper(
    runner: "Runner",
    tx_spendable: str,
    conn_privkey: str = "02",
    global_features: Optional[str] = None,
    features: Optional[str] = None,
) -> List["Event"]:
    """Helper function to make a connection with the node"""
    from lnprototest.utils.bitcoin_utils import tx_spendable
    from lnprototest.stash import stash_field_from_event
    from lnprototest import (
        Connect,
        Block,
        ExpectMsg,
        Msg,
    )

    return [
        Block(blockheight=102, txs=[tx_spendable]),
        Connect(connprivkey=conn_privkey),
        ExpectMsg("init"),
        Msg(
            "init",
            globalfeatures=stash_field_from_event("init", dummy_val="")
            if global_features is None
            else global_features,
            features=stash_field_from_event("init", dummy_val="")
            if features is None
            else features,
        ),
    ]


def open_and_announce_channel_helper(
    runner: "Runner", conn_privkey: str = "02", opts: dict = {}
) -> List["Event"]:
    from lnprototest.utils import gen_random_keyset, pubkey_of
    from lnprototest.utils.bitcoin_utils import (
        BitcoinUtils,
        utxo,
        funding_amount_for_utxo,
    )
    from lnprototest.utils.ln_spec_utils import LightningUtils
    from lnprototest import (
        Block,
        ExpectMsg,
        Msg,
        Commit,
        Side,
        CreateFunding,
        remote_funding_pubkey,
        remote_revocation_basepoint,
        remote_payment_basepoint,
        remote_delayed_payment_basepoint,
        remote_htlc_basepoint,
        remote_per_commitment_point,
        remote_funding_privkey,
        msat,
    )
    from lnprototest.stash import (
        rcvd,
        funding,
        sent,
        commitsig_to_recv,
        channel_id,
        commitsig_to_send,
        funding_txid,
        funding_tx,
    )

    # Make up a channel between nodes 02 and 03, using bitcoin privkeys 10 and 20
    local_keyset = gen_random_keyset()
    local_funding_privkey = "20"
    if "block_height" in opts:
        block_height = opts["block_height"]
    else:
        block_height = 103

    short_channel_id = LightningUtils.derive_short_channel_id(block_height, 1, 0)
    opts["short_channel_id"] = short_channel_id
    opts["block_height"] = block_height + 6
    return [
        Msg(
            "open_channel",
            chain_hash=BitcoinUtils.blockchain_hash(),
            temporary_channel_id="00" * 32,
            funding_satoshis=funding_amount_for_utxo(0),
            push_msat=0,
            dust_limit_satoshis=546,
            max_htlc_value_in_flight_msat=4294967295,
            channel_reserve_satoshis=9998,
            htlc_minimum_msat=0,
            feerate_per_kw=253,
            # clightning uses to_self_delay=6; we use 5 to test differentiation
            to_self_delay=5,
            max_accepted_htlcs=483,
            funding_pubkey=pubkey_of(local_funding_privkey),
            revocation_basepoint=local_keyset.revocation_basepoint(),
            payment_basepoint=local_keyset.payment_basepoint(),
            delayed_payment_basepoint=local_keyset.delayed_payment_basepoint(),
            htlc_basepoint=local_keyset.htlc_basepoint(),
            first_per_commitment_point=local_keyset.per_commit_point(0),
            channel_flags=1,
        ),
        ExpectMsg(
            "accept_channel",
            funding_pubkey=remote_funding_pubkey(),
            revocation_basepoint=remote_revocation_basepoint(),
            payment_basepoint=remote_payment_basepoint(),
            delayed_payment_basepoint=remote_delayed_payment_basepoint(),
            htlc_basepoint=remote_htlc_basepoint(),
            first_per_commitment_point=remote_per_commitment_point(0),
            minimum_depth=3,
            channel_reserve_satoshis=9998,
        ),
        # Create and stash Funding object and FundingTx
        CreateFunding(
            *utxo(0),
            local_node_privkey="02",
            local_funding_privkey=local_funding_privkey,
            remote_node_privkey=runner.get_node_privkey(),
            remote_funding_privkey=remote_funding_privkey(),
        ),
        Commit(
            funding=funding(),
            opener=Side.local,
            local_keyset=local_keyset,
            local_to_self_delay=rcvd("to_self_delay", int),
            remote_to_self_delay=sent("to_self_delay", int),
            local_amount=msat(sent("funding_satoshis", int)),
            remote_amount=0,
            local_dust_limit=546,
            remote_dust_limit=546,
            feerate=253,
            local_features=sent("init.features"),
            remote_features=rcvd("init.features"),
        ),
        Msg(
            "funding_created",
            temporary_channel_id=rcvd(),
            funding_txid=funding_txid(),
            funding_output_index=0,
            signature=commitsig_to_send(),
        ),
        ExpectMsg(
            "funding_signed", channel_id=channel_id(), signature=commitsig_to_recv()
        ),
        # Mine it and get it deep enough to confirm channel.
        Block(blockheight=103, number=3, txs=[funding_tx()]),
        ExpectMsg(
            "channel_ready",
            channel_id=channel_id(),
            second_per_commitment_point=remote_per_commitment_point(1),
        ),
        Msg(
            "channel_ready",
            channel_id=channel_id(),
            second_per_commitment_point=local_keyset.per_commit_point(1),
        ),
        # wait confirmations
        Block(blockheight=103, number=6),
        # FIXME: implement all the utils function for the announcement_signatures
        ExpectMsg(
            "announcement_signatures",
            channel_id=channel_id(),
            short_channel_id=short_channel_id,
            # TODO: How build signatures without hard coding it here?
            node_signature="5ffb05bfb1ef2941cd26e02eea9bfcd6862a08dcfd58473cd1e7da879c2127d6650159c731ae07cd07ff00f4fe7d344aef7997384465f34d7c57add4795a7b09",
            bitcoin_signature="138c93afb2013c39f959e70a163c3d6d8128cf72f8ae143f87b9d1fd6bb0ad30321116b9c58d69fca9fb33c214f681b664e53d5640abc2fdb972dc62a5571053",
        ),
        Msg(
            "announcement_signatures",
            channel_id=channel_id(),
            short_channel_id=short_channel_id,
            # TODO: How build signatures without hard coding it here?
            node_signature="5ffb05bfb1ef2941cd26e02eea9bfcd6862a08dcfd58473cd1e7da879c2127d6650159c731ae07cd07ff00f4fe7d344aef7997384465f34d7c57add4795a7b09",
            bitcoin_signature="138c93afb2013c39f959e70a163c3d6d8128cf72f8ae143f87b9d1fd6bb0ad30321116b9c58d69fca9fb33c214f681b664e53d5640abc2fdb972dc62a5571053",
        ),
    ]



"""stash provides functions which resolve at action time.

During a run, Events often put state in the runner object for later
events to access at action time.  This is called the "stash".

These accessors actually return functions, which most Events know
means they're to be called at action time.  For the sake of (perhaps
non-Pythony) test authors, they are always called as (), even if they
simply return another function.

"""
from .stash import (
    commitsig_to_send,
    commitsig_to_recv,
    channel_id,
    channel_announcement,
    channel_update,
    get_member,
    rcvd,
    sent,
    funding_amount,
    funding_pubkey,
    funding_tx,
    funding_txid,
    funding,
    funding_close_tx,
    htlc_sigs_to_send,
    htlc_sigs_to_recv,
    locking_script,
    witnesses,
    stash_field_from_event,
)



"""
Bitcoin utils is a collection of methods that helps to
work with bitcoin primitive.
"""
import hashlib

from enum import Enum
from typing import Tuple

import bitcoin.core

from bitcoin.core import Hash160, x
from bitcoin.core.script import OP_0, OP_CHECKSIG, CScript
from bitcoin.wallet import CBitcoinSecret


class ScriptType(Enum):
    """
    Type of Script used in the Runner.

    In particular, during the testing we need to have
    two type of script, the valid one and the invalid one.
    This is useful when is needed to send an invalid script.

    FIXME: naming is too simple.
    """

    VALID_CLOSE_SCRIPT = 1
    INVALID_CLOSE_SCRIPT = 2


class BitcoinUtils:
    """Main implementation class of the lightning networks utils.

    The implementation class contains only static methods that
    apply the rules specified by the BIP."""

    @staticmethod
    def blockchain_hash() -> str:
        """Return the chain transaction hash.
        That in this case is the regtest transaction hash."""
        return "06226e46111a0b59caaf126043eb5bbf28c34f3a5e332a1fc7b2b73cf188910f"

    @staticmethod
    def build_valid_script(
        script_type: ScriptType = ScriptType.VALID_CLOSE_SCRIPT,
        word: str = "lnprototest",
    ) -> str:
        """Build a valid bitcoin script and hide the primitive of the library"""
        secret_str = f"correct horse battery staple {word}"
        h = hashlib.sha256(secret_str.encode("ascii")).digest()
        seckey = CBitcoinSecret.from_secret_bytes(h)
        if script_type is ScriptType.VALID_CLOSE_SCRIPT:
            return CScript([OP_0, Hash160(seckey.pub)]).hex()
        elif script_type is ScriptType.INVALID_CLOSE_SCRIPT:
            return CScript([seckey.pub, OP_CHECKSIG]).hex()

    @staticmethod
    def build_script(hex: str) -> CScript:
        return CScript(x(hex))


# Here are the keys to spend funds, derived from BIP32 seed
# `0000000000000000000000000000000000000000000000000000000000000001`:
#
#    pubkey 0/0/1: 02d6a3c2d0cf7904ab6af54d7c959435a452b24a63194e1c4e7c337d3ebbb3017b
#    privkey 0/0/1: 76edf0c303b9e692da9cb491abedef46ca5b81d32f102eb4648461b239cb0f99
#    WIF 0/0/1: cRZtHFwyrV3CS1Muc9k4sXQRDhqA1Usgi8r7NhdEXLgM5CUEZufg
#    P2WPKH 0/0/1: bcrt1qsdzqt93xsyewdjvagndw9523m27e52er5ca7hm
#    UTXO: d3fb780146954eb42e371c80cbee1725f8ae330848522f105bda24e1fb1fc010/1 (0.01BTC)
#
#    pubkey 0/0/2: 038f1573b4238a986470d250ce87c7a91257b6ba3baf2a0b14380c4e1e532c209d
#    privkey 0/0/2: bc2f48a76a6b8815940accaf01981d3b6347a68fbe844f81c50ecbadf27cd179
#    WIF 0/0/2: cTtWRYC39drNzaANPzDrgoYsMgs5LkfE5USKH9Kr9ySpEEdjYt3E
#    P2WPKH 0/0/2: bcrt1qlkt93775wmf33uacykc49v2j4tayn0yj25msjn
#    UTXO: d3fb780146954eb42e371c80cbee1725f8ae330848522f105bda24e1fb1fc010/0 (0.02BTC)
#
#    pubkey 0/0/3: 02ffef0c295cf7ca3a4ceb8208534e61edf44c606e7990287f389f1ea055a1231c
#    privkey 0/0/3: 16c5027616e940d1e72b4c172557b3b799a93c0582f924441174ea556aadd01c
#    WIF 0/0/3: cNLxnoJSQDRzXnGPr4ihhy2oQqRBTjdUAM23fHLHbZ2pBsNbqMwb
#    P2WPKH 0/0/3: bcrt1q2ng546gs0ylfxrvwx0fauzcvhuz655en4kwe2c
#    UTXO: d3fb780146954eb42e371c80cbee1725f8ae330848522f105bda24e1fb1fc010/3 (0.03BTC)
#
#    pubkey 0/0/4: 026957e53b46df017bd6460681d068e1d23a7b027de398272d0b15f59b78d060a9
#    privkey 0/0/4: 53ac43309b75d9b86bef32c5bbc99c500910b64f9ae089667c870c2cc69e17a4
#    WIF 0/0/4: cQPMJRjxse9i1jDeCo8H3khUMHYfXYomKbwF5zUqdPrFT6AmtTbd
#    P2WPKH 0/0/4: bcrt1qrdpwrlrmrnvn535l5eldt64lxm8r2nwkv0ruxq
#    UTXO: d3fb780146954eb42e371c80cbee1725f8ae330848522f105bda24e1fb1fc010/4 (0.04BTC)
#
#    pubkey 0/0/5: 03a9f795ff2e4c27091f40e8f8277301824d1c3dfa6b0204aa92347314e41b1033
#    privkey 0/0/5: 16be98a5d4156f6f3af99205e9bc1395397bca53db967e50427583c94271d27f
#    WIF 0/0/5: cNLuxyjvR6ga2q6fdmSKxAd1CPQDShKV9yoA7zFKT7GJwZXr9MmT
#    P2WPKH 0/0/5: bcrt1q622lwmdzxxterumd746eu3d3t40pq53p62zhlz
#    UTXO: d3fb780146954eb42e371c80cbee1725f8ae330848522f105bda24e1fb1fc010/2 (48.89994700BTC)
#
#
# We add another UTXO which is solely spendable by the test framework, and not accessible to the
# runner -- needed for dual-funded tests.
#
#    pubkey: 02c6047f9441ed7d6d3045406e95c07cd85c778e4b8cef3ca7abac09b95c709ee5
#    privkey: 0000000000000000000000000000000000000000000000000000000000000002
#    P2WPKH : bcrt1qq6hag67dl53wl99vzg42z8eyzfz2xlkvwk6f7m
#    UTXO: d3fb780146954eb42e371c80cbee1725f8ae330848522f105bda24e1fb1fc010/5 (0.005BTC)
#
# UTXO with a 1-of-5 multisig (results in a long/expensive witness)
# -- needed for dual-funded tests.
#
#   pubkey_1: 0253cdf835e328346a4f19de099cf3d42d4a7041e073cd4057a1c4fd7cdbb1228f
#   privkey_1: cPToYwmZxXaAaAw6XHi8aEVefYAoYxPs8TCFpsAF6JVWbg1NSaZE
#   pubkey_2: 03ae903722f21f85e651b8f9b18fc854084fb90eeb76452bdcfd0cb43a16a382a2
#   privkey_2: cMsCoM96my8y5DDHLJfpGzFo7EhnhNFscbARtNS8mz7vTV5PiVR3
#   pubkey_3: 036c264d68a9727afdc75949f7d7fa71910ae9ae8001a1fbffa6f7ce000976597c
#   privkey_3: cQE1RbQyGcecfzTurwayy9vJtapRmM9isUG6AVTxYf2seiBHcNMp
#   pubkey_4: 036429fa8a4ef0b2b1d5cb553e34eeb90a32ab19fae1f0024f332ab4f74283a728
#   privkey_4: cVKnBoJ294xQyzRWFkxSmxMr1PFdHxGgTVBrMWpSF6o8KoJhWbNY
#   pubkey_5: 03d4232f19ea85051e7b76bf5f01d03e17eea8751463dee36d71413a739de1a927
#   privkey_5: cRboxysVFZUxZQ3DdvoTRuBKfStQ46MAA5HTVgVQW6VKF88HrC4H
#   script: 51210253cdf835e328346a4f19de099cf3d42d4a7041e073cd4057a1c4fd7cdbb1228f2103ae903722f21f85e651b8f9b18fc854084fb90eeb76452bdcfd0cb43a16a382a221036c264d68a9727afdc75949f7d7fa71910ae9ae8001a1fbffa6f7ce000976597c21036429fa8a4ef0b2b1d5cb553e34eeb90a32ab19fae1f0024f332ab4f74283a7282103d4232f19ea85051e7b76bf5f01d03e17eea8751463dee36d71413a739de1a92755ae
#   P2WSH: bcrt1qug62lyrfd7khs7welgu28y66zzuq5nc4t9gdnyx3rjm9fud2f7gqm0ksxn
#   UTXO: d3fb780146954eb42e371c80cbee1725f8ae330848522f105bda24e1fb1fc010/6 (0.06BTC)
#
tx_spendable = "0200000000010184591a56720aabc8023cecf71801c5e0f9d049d0c550ab42412ad12a67d89f3a0000000000feffffff0780841e0000000000160014fd9658fbd476d318f3b825b152b152aafa49bc9240420f000000000016001483440596268132e6c99d44dae2d151dabd9a2b232c180a2901000000160014d295f76da2319791f36df5759e45b15d5e105221c0c62d000000000016001454d14ae910793e930d8e33d3de0b0cbf05aa533300093d00000000001600141b42e1fc7b1cd93a469fa67ed5eabf36ce354dd620a107000000000016001406afd46bcdfd22ef94ac122aa11f241244a37ecc808d5b000000000022002000b068df6e0e0542e776cea5ebe8f5f1a9b40b531ddd8e94b1a7ff9829b5bbaa024730440220367b9bfed0565bad2137124f736373626fa3135e59b20a7b5c1d8f2b8f1b26bb02202f664de39787082a376d222487f02ef19e45696c041044a6d579eecabb68e94501210356609a904a7026c7391d3fbf71ad92a00e04b4cd2fb6a8d1e69cbc0998f6690a65000000"


def utxo(index: int = 0) -> Tuple[str, int, int, str, int]:
    """Helper to get a P2WPKH UTXO, amount, privkey and fee from the tx_spendable transaction"""

    amount = (index + 1) * 1000000
    if index == 0:
        txout = 1
        key = "76edf0c303b9e692da9cb491abedef46ca5b81d32f102eb4648461b239cb0f99"
    elif index == 1:
        txout = 0
        key = "bc2f48a76a6b8815940accaf01981d3b6347a68fbe844f81c50ecbadf27cd179"
    elif index == 2:
        txout = 3
        key = "16c5027616e940d1e72b4c172557b3b799a93c0582f924441174ea556aadd01c"
    elif index == 3:
        txout = 4
        key = "53ac43309b75d9b86bef32c5bbc99c500910b64f9ae089667c870c2cc69e17a4"
    elif index == 4:
        txout = 2
        key = "16be98a5d4156f6f3af99205e9bc1395397bca53db967e50427583c94271d27f"
        amount = 4983494700
    elif index == 5:
        txout = 5
        key = "0000000000000000000000000000000000000000000000000000000000000002"
        amount = 500000
    elif index == 6:
        txout = 6
        key = "38204720bc4f9647fd58c6d0a4bd3a6dd2be16d8e4273c4d1bdd5774e8c51eaf"
        amount = 6000000
    else:
        raise ValueError("index must be 0-6 inclusive")

    # Reasonable funding fee in sats
    reasonable_funding_fee = 200

    return txid_raw(tx_spendable), txout, amount, key, reasonable_funding_fee


def tx_out_for_index(index: int = 0) -> int:
    _, txout, _, _, _ = utxo(index)
    return txout


def utxo_amount(index: int = 0) -> int:
    """How much is this utxo worth"""
    _, _, amt, _, _ = utxo(index)
    return amt


def funding_amount_for_utxo(index: int = 0) -> int:
    """How much can we fund a channel for using utxo #index?"""
    _, _, amt, _, fee = utxo(index)
    return amt - fee


def txid_raw(tx: str) -> str:
    """Helper to get the txid of a tx: note this is in wire protocol order, not bitcoin order!"""
    return bitcoin.core.CTransaction.deserialize(bytes.fromhex(tx)).GetTxid().hex()



from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from .event import Event


class EventError(Exception):
    """Error thrown when the runner fails in some way"""

    def __init__(self, event: "Event", message: str):
        self.eventpath = [event]
        self.message = message

    def add_path(self, event: "Event") -> None:
        self.eventpath = [event] + self.eventpath

    def path_to_str(self) -> str:
        result = "["
        for event in self.eventpath:
            result += f"{event},"
        return f"{result}]"

    def __str__(self) -> str:
        return f"`{self.message}` on event {self.path_to_str()}"


class SpecFileError(Exception):
    """Error thrown when the specification file has an error"""

    def __init__(self, event: "Event", message: str):
        self.event = event
        self.message = message



#! /usr/bin/python3
import pyln.spec.bolt1
import pyln.spec.bolt2
import pyln.spec.bolt7
from pyln.proto.message import MessageNamespace
from .signature import SigType
from typing import List


def make_namespace(csv: List[str]) -> MessageNamespace:
    """Load a namespace, replacing signature type"""
    ns = MessageNamespace()
    # We replace the fundamental signature type with our custom type,
    # then we load in all the csv files so they use it.
    ns.fundamentaltypes["signature"] = SigType()
    ns.load_csv(csv)
    return ns


def peer_message_namespace() -> MessageNamespace:
    """Namespace containing all the peer messages"""
    return make_namespace(
        pyln.spec.bolt1.csv + pyln.spec.bolt2.csv + pyln.spec.bolt7.csv
    )


def namespace() -> MessageNamespace:
    return event_namespace


def assign_namespace(ns: MessageNamespace) -> None:
    global event_namespace
    event_namespace = ns


# By default, include all peer message bolts
event_namespace = peer_message_namespace()



#! /usr/bin/python3
import logging
import shutil
import tempfile

import coincurve
import functools

from .errors import SpecFileError
from .structure import Sequence
from .event import Event, MustNotMsg, ExpectMsg
from .utils import privkey_expand
from .keyset import KeySet
from abc import ABC, abstractmethod
from typing import Dict, Optional, List, Union, Any, Callable


class Conn(object):
    """Class for connections.  Details filled in by the particular runner."""

    def __init__(self, connprivkey: str):
        """Create a connection from a node with the given hex privkey: we use
        trivial values for private keys, so we simply left-pad with zeroes"""
        self.name = connprivkey
        self.connprivkey = privkey_expand(connprivkey)
        self.pubkey = coincurve.PublicKey.from_secret(self.connprivkey.secret)
        self.expected_error = False
        self.must_not_events: List[MustNotMsg] = []

    def __str__(self) -> str:
        return self.name


class Runner(ABC):
    """Abstract base class for runners.

    Most of the runner parameters can be extracted at runtime, but we do
    require that minimum_depth be 3, just for test simplicity.
    """

    def __init__(self, config: Any):
        self.config = config
        self.directory = tempfile.mkdtemp(prefix="lnpt-cl-")
        # key == connprivkey, value == Conn
        self.conns: Dict[str, Conn] = {}
        self.last_conn: Optional[Conn] = None
        self.stash: Dict[str, Dict[str, Any]] = {}
        self.logger = logging.getLogger(__name__)
        if self.config.getoption("verbose"):
            self.logger.setLevel(logging.DEBUG)
        else:
            self.logger.setLevel(logging.INFO)

    def _is_dummy(self) -> bool:
        """The DummyRunner returns True here, as it can't do some things"""
        return False

    def find_conn(self, connprivkey: Optional[str]) -> Optional[Conn]:
        # Default is whatever we specified last.
        if connprivkey is None:
            return self.last_conn
        if connprivkey in self.conns:
            self.last_conn = self.conns[connprivkey]
            return self.last_conn
        return None

    def add_conn(self, conn: Conn) -> None:
        self.conns[conn.name] = conn
        self.last_conn = conn

    def disconnect(self, event: Event, conn: Conn) -> None:
        if conn is None:
            raise SpecFileError(event, "Unknown conn")
        del self.conns[conn.name]
        self.check_final_error(event, conn, conn.expected_error, conn.must_not_events)

    def check_error(self, event: Event, conn: Conn) -> Optional[str]:
        conn.expected_error = True
        return None

    def post_check(self, sequence: Sequence) -> None:
        # Make sure no connection had an error.
        for conn_name in list(self.conns.keys()):
            logging.debug(
                f"disconnection connection with key={conn_name} and value={self.conns[conn_name]}"
            )
            self.disconnect(sequence, self.conns[conn_name])

    def restart(self) -> None:
        self.conns = {}
        self.last_conn = None
        self.stash = {}

    # FIXME: Why can't we use SequenceUnion here?
    def run(self, events: Union[Sequence, List[Event], Event]) -> None:
        sequence = Sequence(events)
        self.start()
        while True:
            all_done = sequence.action(self)
            self.post_check(sequence)
            if all_done:
                self.stop()
                return
            self.restart()

    def add_stash(self, stashname: str, vals: Any) -> None:
        """Add a dict to the stash."""
        self.stash[stashname] = vals

    def get_stash(self, event: Event, stashname: str, default: Any = None) -> Any:
        """Get an entry from the stash."""
        if stashname not in self.stash:
            if default is not None:
                return default
            raise SpecFileError(event, "Unknown stash name {}".format(stashname))
        return self.stash[stashname]

    def teardown(self):
        """The Teardown method is called at the end of the test,
        and it is used to clean up the root dir where the tests are run."""
        shutil.rmtree(self.directory)

    @abstractmethod
    def is_running(self) -> bool:
        """Return a boolean value that tells whether the runner is running
        or not.
        Is leave up to the runner implementation to keep the runner state"""
        pass

    @abstractmethod
    def connect(self, event: Event, connprivkey: str) -> None:
        pass

    @abstractmethod
    def check_final_error(
        self,
        event: Event,
        conn: Conn,
        expected: bool,
        must_not_events: List[MustNotMsg],
    ) -> None:
        pass

    @abstractmethod
    def start(self) -> None:
        pass

    @abstractmethod
    def stop(self, print_logs: bool = False) -> None:
        """
        Stop the runner, and print all the log that the ln
        implementation produced.

        Print the log is useful when we have a failure e we need
        to debug what happens during the tests.
        """
        pass

    @abstractmethod
    def recv(self, event: Event, conn: Conn, outbuf: bytes) -> None:
        pass

    @abstractmethod
    def get_output_message(self, conn: Conn, event: ExpectMsg) -> Optional[bytes]:
        pass

    @abstractmethod
    def getblockheight(self) -> int:
        pass

    @abstractmethod
    def trim_blocks(self, newheight: int) -> None:
        pass

    @abstractmethod
    def add_blocks(self, event: Event, txs: List[str], n: int) -> None:
        pass

    @abstractmethod
    def expect_tx(self, event: Event, txid: str) -> None:
        pass

    @abstractmethod
    def invoice(self, event: Event, amount: int, preimage: str) -> None:
        pass

    @abstractmethod
    def accept_add_fund(self, event: Event) -> None:
        pass

    @abstractmethod
    def fundchannel(
        self,
        event: Event,
        conn: Conn,
        amount: int,
        feerate: int = 0,
        expect_fail: bool = False,
    ) -> None:
        pass

    @abstractmethod
    def init_rbf(
        self,
        event: Event,
        conn: Conn,
        channel_id: str,
        amount: int,
        utxo_txid: str,
        utxo_outnum: int,
        feerate: int,
    ) -> None:
        pass

    @abstractmethod
    def addhtlc(self, event: Event, conn: Conn, amount: int, preimage: str) -> None:
        pass

    @abstractmethod
    def get_keyset(self) -> KeySet:
        pass

    @abstractmethod
    def get_node_privkey(self) -> str:
        pass

    @abstractmethod
    def get_node_bitcoinkey(self) -> str:
        pass

    @abstractmethod
    def has_option(self, optname: str) -> Optional[str]:
        pass

    @abstractmethod
    def add_startup_flag(self, flag: str) -> None:
        pass

    @abstractmethod
    def close_channel(self, channel_id: str) -> None:
        """
        Close the channel with the specified channel id.

        :param channel_id:  the channel id as string value where the
        caller want to close;
        :return No value in case of success is expected,
        but an `RpcError` is expected in case of err.
        """
        pass


def remote_revocation_basepoint() -> Callable[[Runner, Event, str], str]:
    """Get the remote revocation basepoint"""

    def _remote_revocation_basepoint(runner: Runner, event: Event, field: str) -> str:
        return runner.get_keyset().revocation_basepoint()

    return _remote_revocation_basepoint


def remote_payment_basepoint() -> Callable[[Runner, Event, str], str]:
    """Get the remote payment basepoint"""

    def _remote_payment_basepoint(runner: Runner, event: Event, field: str) -> str:
        return runner.get_keyset().payment_basepoint()

    return _remote_payment_basepoint


def remote_delayed_payment_basepoint() -> Callable[[Runner, Event, str], str]:
    """Get the remote delayed_payment basepoint"""

    def _remote_delayed_payment_basepoint(
        runner: Runner, event: Event, field: str
    ) -> str:
        return runner.get_keyset().delayed_payment_basepoint()

    return _remote_delayed_payment_basepoint


def remote_htlc_basepoint() -> Callable[[Runner, Event, str], str]:
    """Get the remote htlc basepoint"""

    def _remote_htlc_basepoint(runner: Runner, event: Event, field: str) -> str:
        return runner.get_keyset().htlc_basepoint()

    return _remote_htlc_basepoint


def remote_funding_pubkey() -> Callable[[Runner, Event, str], str]:
    """Get the remote funding pubkey (FIXME: we assume there's only one!)"""

    def _remote_funding_pubkey(runner: Runner, event: Event, field: str) -> str:
        return (
            coincurve.PublicKey.from_secret(
                privkey_expand(runner.get_node_bitcoinkey()).secret
            )
            .format()
            .hex()
        )

    return _remote_funding_pubkey


def remote_funding_privkey() -> Callable[[Runner, Event, str], str]:
    """Get the remote funding privkey (FIXME: we assume there's only one!)"""

    def _remote_funding_privkey(runner: Runner, event: Event, field: str) -> str:
        return runner.get_node_bitcoinkey()

    return _remote_funding_privkey


def remote_per_commitment_point(n: int) -> Callable[[Runner, Event, str], str]:
    """Get the n'th remote per-commitment point"""

    def _remote_per_commitment_point(
        n: int, runner: Runner, event: Event, field: str
    ) -> str:
        return runner.get_keyset().per_commit_point(n)

    return functools.partial(_remote_per_commitment_point, n)


def remote_per_commitment_secret(n: int) -> Callable[[Runner, Event, str], str]:
    """Get the n'th remote per-commitment secret"""

    def _remote_per_commitment_secret(runner: Runner, event: Event, field: str) -> str:
        return runner.get_keyset().per_commit_secret(n)

    return _remote_per_commitment_secret



#! /usr/bin/python3
import coincurve
from io import BufferedIOBase
from pyln.proto.message import FieldType, split_field
from .utils import check_hex, privkey_expand
from typing import Union, Tuple, Dict, Any, Optional, cast


class Sig(object):
    """The value of a signature, either as a privkey/hash pair or a raw
    signature.  This has the property that if the raw signature is a valid
    signature of privkey over hash, they are considered "equal"
    """

    def __init__(self, *args: Any):
        """Either a 64-byte hex/bytes value, or a PrivateKey and a hash"""
        if len(args) == 1:
            if type(args[0]) is bytes:
                if len(args[0]) != 64:
                    raise ValueError("Sig() with 1 arg expects 64 bytes or 128 hexstr")
                self.sigval: Union[bytes, None] = cast(bytes, args[0])
            else:
                if type(args[0]) is not str:
                    raise TypeError("Expected hexsig or Privkey, hash")
                if len(args[0]) != 128:
                    self.sigval = Sig.from_der(bytes.fromhex(args[0]))
                else:
                    self.sigval = bytes.fromhex(args[0])
        elif len(args) == 2:
            self.sigval = None
            self.privkey = privkey_expand(args[0])
            self.hashval = bytes.fromhex(check_hex(args[1], 64))
        else:
            raise TypeError("Expected hexsig or Privkey, hash")

    @staticmethod
    def to_der(b: bytes) -> bytes:
        """Seriously fuck off with DER encoding :("""
        r = b[0:32]
        s = b[32:64]
        # Trim zero bytes
        while r[0] == 0:
            r = r[1:]
        # Prepend 0 again if would be negative
        if r[0] & 0x80:
            r = bytes([0]) + r
        # Trim zero bytes
        while s[0] == 0:
            s = s[1:]
        # Prepend 0 again if would be negative
        if s[0] & 0x80:
            s = bytes([0]) + s

        # 2 == integer, next == length
        ret = bytes([0x02, len(r)]) + r + bytes([0x02, len(s)]) + s
        # 30 == compound, next = length
        return bytes([0x30, len(ret)]) + ret

    @staticmethod
    def from_der(b: bytes) -> bytes:
        """Sigh.  Seriously, WTF is it with DER encoding?"""
        if b[0] != 0x30 or b[1] != len(b) - 2 or b[2] != 0x02:
            raise ValueError("{} is not a DER signature?".format(b.hex()))
        rlen = b[3]
        r = b[4 : 4 + rlen].rjust(32, bytes(1))[-32:]
        assert len(r) == 32
        if b[4 + rlen] != 0x02:
            raise ValueError("{} is not a DER signature?".format(b.hex()))
        s = b[4 + rlen + 1 + 1 :].rjust(32, bytes(1))[-32:]
        assert len(s) == 32
        return r + s

    def __eq__(self, other: Any) -> bool:
        # For convenience of using stashed objects, we allow comparison with str
        if isinstance(other, str):
            othersig = Sig(other)
        else:
            othersig = cast(Sig, other)
        if self.sigval and othersig.sigval:
            return self.sigval == othersig.sigval
        elif not self.sigval and not othersig.sigval:
            return self.privkey == othersig.privkey and self.hashval == othersig.hashval
        elif not self.sigval:
            a = self
            b = othersig
        else:
            a = othersig
            b = self
        # A has a privkey/hash, B has a sigval.
        pubkey = coincurve.PublicKey.from_secret(a.privkey.secret)
        assert b.sigval is not None
        if coincurve.verify_signature(
            self.to_der(b.sigval), a.hashval, pubkey.format(), hasher=None
        ):
            return True
        return False

    def to_str(self) -> str:
        if self.sigval:
            return self.sigval.hex()
        else:
            return "Sig({},{})".format(self.privkey.secret.hex(), self.hashval.hex())

    @staticmethod
    def from_str(s: str) -> Tuple["Sig", str]:
        a, b = split_field(s)
        if a.startswith("Sig("):
            privkey = a[4:]
            a, b = split_field(b[1:])
            # Trim ) off Sig()
            return Sig(privkey, a[:-1]), b
        return Sig(bytes.fromhex(a)), b

    def to_bin(self) -> bytes:
        if not self.sigval:
            return self.from_der(self.privkey.sign(self.hashval, hasher=None))
        else:
            return self.sigval


class SigType(FieldType):
    """A signature type which has special comparison properties"""

    def __init__(self) -> None:
        super().__init__("signature")

    def val_to_str(self, v: Sig, otherfields: Dict[str, Any]) -> str:
        return v.to_str()

    def val_from_str(self, s: str) -> Tuple["Sig", str]:
        return Sig.from_str(s)

    def write(
        self, io_out: BufferedIOBase, v: Sig, otherfields: Dict[str, Any]
    ) -> None:
        io_out.write(v.to_bin())

    def read(self, io_in: BufferedIOBase, otherfields: Dict[str, Any]) -> Optional[Sig]:
        val = io_in.read(64)
        if len(val) == 0:
            return None
        elif len(val) != 64:
            raise ValueError("{}: not enough remaining".format(self))
        return Sig(val)


def test_der() -> None:
    der = b"0E\x02!\x00\xa0\xb3\x7f\x8f\xbah<\xc6\x8fet\xcdC\xb3\x9f\x03C\xa5\x00\x08\xbfl\xce\xa9\xd121\xd9\xe7\xe2\xe1\xe4\x02 \x11\xed\xc8\xd3\x07%B\x96&J\xeb\xfc=\xc7l\xd8\xb6h7:\x07/\xd6Fe\xb5\x00\x00\xe9\xfc\xceR"
    sig = Sig.from_der(der)
    der2 = Sig.to_der(sig)
    assert der == der2


def test_signature() -> None:
    s = Sig("01", "00" * 32)

    assert s == s
    b = s.to_bin()
    s2 = Sig(b)

    assert s == s2
    assert s2 == s
